/* Last edited on 2004-08-25 16:17:02 by stolfi */
/* Tests relative speed of quicksort and heapsort */

#include <stdio.h>
#include <stdlib.h>
#include <js.h>
#include <sortheap.h>
#include <math.h>

static int ncmp;
static long int *data; /* Random values to be compared. */
static int ndata; /* Size of {data} vector. */

#define NITS 50
#define NMAX (256*256)

typedef int tsr_cmp_t(int a, int b); 
  /* A signed comparison predicate. */

typedef void tsr_sorter_t(int *h, int n, tsr_cmp_t cmp, int sgn);
  /* A procedure that sorts {h} so that 
    {sgn*cmp(h[i-1],h[i]) <= 0} for all {i} */
    
/* PROTOS */

int tsr_test_sorter(int *h, int n, tsr_sorter_t srt, tsr_cmp_t cmp, int sgn);
  /* Tests sorter {srt} on {h} with comparator {sgn*cmp}; checks
    order of result, returns comparison count. */
    
void tsr_print_stats(FILE *wr, char *sname, int n, int *nc, int nits);
  /* Prints data for sorter {sname} on an array of {n} elements,
    given the comparison counts {nc[0..nits-1]} obtained in {nits}
    test runs. */

int datacmp(int a, int b);
  /* Compares {data[a]} with {data[b]}. */
  
void tsr_heap_sort(int *h, int n, tsr_cmp_t cmp, int sgn);
void tsr_quick_sort(int *h, int n, tsr_cmp_t cmp, int sgn);
void tsr_whop_sort(int *h, int n, tsr_cmp_t cmp, int sgn);
  /* Three competing sorters. */

int main (int argc, char **argv)
  {
    int n = 0;
    fprintf(stdout, "  %-10s & %6s & %7s & %7s & %10s & %10s \\\\\n",
      "Algorithm", "n", "min", "max", "avg", "std"); 
    while (n <= NMAX) 
      { /* Allocate data array */
        long int dt[n]; data = dt; ndata = n;
        int h[n];
        int hc[NITS]; /* Comparison counts for heapsort. */
        int zc[NITS]; /* Comparison counts for whopsort. */
        int jc[NITS]; /* Comparison counts for {sortheap} of {libjs.a}. */
        int qc[NITS]; /* Comparison counts for quicksort. */
        int sgn = +1;

        int it;
        for (it = 0; it < NITS; it++)
          { /* Generate "random" numbers and test the two sorters on them: */
            int i;
            srandom(it+1);
            for (i = 0; i < n; i++) { dt[i] = random(); }
            
            hc[it] = tsr_test_sorter(h, n, tsr_heap_sort, datacmp, sgn);
            for (i = 0; i < n; i++) { h[i] = i; }
            zc[it] = tsr_test_sorter(h, n, tsr_whop_sort, datacmp, sgn);
            for (i = 0; i < n; i++) { h[i] = i; }
            jc[it] = tsr_test_sorter(h, n, shp_heap_sort, datacmp, sgn);
            for (i = 0; i < n; i++) { h[i] = i; }
            qc[it] = tsr_test_sorter(h, n, tsr_quick_sort, datacmp, sgn);
          }
        tsr_print_stats(stdout, "standard",  n, hc, NITS);
        tsr_print_stats(stdout, "modified",  n, zc, NITS);
        tsr_print_stats(stdout, "jslib",     n, jc, NITS);
        tsr_print_stats(stdout, "quicksort", n, qc, NITS);
        fprintf(stdout, "\\hline\n");

        n = (n == 0 ? 1 : 2*n); 
      }
    return 0;
  }

void tsr_print_stats(FILE *wr, char *sname, int n, int *nc, int nits)
  {
    fprintf(wr, "  %-10s & %6d", sname, n);
    double tc = 0;
    int it;
    /* Compute min and max cmps: */
    int cmin, cmax;
    cmin = cmax = nc[0];
    for (it = 1; it < nits; it++)
      { int c = nc[it]; 
        if (c < cmin) 
          { cmin = c; }
        else if
          (c > cmax)
        { cmax = c; }
      }
    fprintf(wr, " & %8d & %8d", cmin, cmax);
    /* Compute average cmps: */
    for (it = 0; it < nits; it++)
      { int c = nc[it]; 
        tc += c;
      }
    double avg = tc/nits;
    /* Compute standard deviation of cmps: */
    double tcc = 0;
    for (it = 0; it < nits; it++)
      { double dc = nc[it] - avg; 
        tcc += dc*dc;
      }
    double var = tcc/(nits-1);
    double std = (var < 0.0 ? 0.0 : sqrt(var));
    /* Print them out: */
    fprintf(wr, " & %10.1f & %10.1f \\\\\n", avg, std);
    fflush(wr);
  }

int tsr_test_sorter(int *h, int n, tsr_sorter_t srt, tsr_cmp_t cmp, int sgn)
  { 
    ncmp = 0;
    /* Start with trivial perm */
    int i;
    for (i = 0; i < n; i++) { h[i] = i; }
    /* Sort the data: */
    srt(h, n, cmp, sgn);
    int nc = ncmp;
    /* Check range: */
    for (i = 0; i < n; i++)
      { int hi = h[i]; 
        if ((hi < 0) || (hi >= n)) {  assert(FALSE, "not 0..n-1"); }
      }
    /* Check ordering: */
    for (i = 1; i < n; i++)
      { if (sgn*cmp(h[i-1],h[i]) > 0) 
          { assert(FALSE, "out of order"); }
      }
    /* Check permutation: */
    for (i = 0; i < n; i++)
      { int hi = h[i]; 
        if (hi < 0) { hi = -hi-1; }
        if (h[hi] < 0) {  assert(FALSE, "not perm"); }
        h[hi] = -h[hi]-1;
      }
    return nc;
  }

int datacmp(int a, int b)
  {
    assert((a >= 0) && (a < ndata), "datacmp: bad a");
    assert((b >= 0) && (b < ndata), "datacmp: bad b");
    ncmp++;
    long int da = data[a], db = data[b];
    if (da < db) 
      { return -1; }
    else if (da > db)
      { return +1; }
    else
      { return 0; }
  }

void tsr_heap_sort(int *h, int n, tsr_cmp_t cmp, int sgn)
  { /* Arrange elements into a sorted heap, in reverse order: */
    int m = 0;
    while (m < n) 
      { /* Grab next element {v}: */
        int v = h[m];
        /* Create a vacancy {h[i]} at the end of the heap: */
        int i = m, j;
        m++;
        /* Bubble it rootwards to its proper place {h[i]}: */
        while ((i > 0) && (sgn*cmp(v,h[j=(i-1)/2]) > 0)) { h[i] = h[j]; i = j; }
        h[i] = v;
      }
    /* Pop them out in reverse order and store them at the end: */
    while (m > 1) 
      { /* Swap root with last element, decrement heap: */
        m--; int w = h[m]; h[m] = h[0]; h[0] = w;
        /* Bubble the new root down to its proper place: */
        int j = 0;
        int ia = 1; /* {h[ia]} is the first child of {h[j]}. */
        while (ia < m)
          { /* Find largest child {h[i]} of {h[j]}: */
            int ib = ia + 1; /*  {h[ib]} is the second child of {h[j]}. */
            int i = ((ib < m) && (sgn*cmp(h[ia],h[ib]) < 0) ? ib : ia);
            /* Ensure that {w=h[j]} precedes the smallest child: */
            if (sgn*cmp(w,h[i]) < 0)
              { /* Swap {h[j]} with child {h[i]}: */
                h[j] = h[i]; h[i] = w; 
                j = i;
              }
            else
              { /* Item {h[j]} is correctly placed, stop: */
                j = m;
              }
            ia = 2*j + 1;
          }
      }
  }

#define SMALL 5

void tsr_quick_sort(int *h, int n, tsr_cmp_t cmp, int sgn)
  {
    auto void do_qsort(int i, int j); 
      /* Sorts segment {h[i..j-1]} */

    void do_qsort(int i, int j)
      {
        while (j-i > SMALL)
          { /* Select separator element: */
            int m = (i+j-1)/2;
            int v = h[m];
            /* Put the first element h[i] in its place: */
            h[m] = h[i];
            /* Sift {h[i+1..j-1]} into leqs {h[i+1..r-1]} and geqs {h[s+1..j-1]}: */
            int r = i+1; int s = j-1;
            do
              { while ((r <= s) && (sgn*cmp(h[r],v) <= 0)) { r++; }
                while ((s >= r) && (sgn*cmp(v,h[s]) <= 0)) { s--; }
                if (r <= s)
                  { /* Note {r == s} is impossible here. */
                    assert(r < s, "program error");
                    int t = h[r]; h[r] = h[s]; h[s] = t; r++; s--;
                  }
              }
            while (r <= s);
            /* Now leqs = {h[i+1..s]}, geqs = {h[r..j-1]}. */
            /* Insert {v} back between leqs and geqs: */
            h[i] = h[s]; h[s] = v;
            /* Now leqs = {h[i..s-1]}, geqs = {h[r..j-1]}. */
            /* Recurse on smaller half, iterate on larger one: */
            if (s - i < j - r)
              { if (s - i >= 2) { do_qsort(i, s); } i = r; }
            else
              { if (j - r >= 2) { do_qsort(r, j); } j = s; }
          }
        /* Finish with insertion sort */
        int k = i+1;
        while (k < j)
          { int v = h[k]; int r = k;
            while ((r > i) && (sgn*cmp(h[r-1],v) > 0)) { h[r] = h[r-1]; r--; }
            h[r] = v;
            k++;
          }
      }
    do_qsort(0, n);
  }

void tsr_whop_sort(int *h, int n, tsr_cmp_t cmp, int sgn)
  { /* Arrange elements into a sorted whop, in reverse order: */
    int m = 0;
    while (m < n) 
      { /* Insert at bottom of heap: */
        int i = m;
        m++;
        /* Grab dubious element {v}: */
        int v = h[i], j;
        /* Bubble it rootwards to its proper place {h[i]}: */
        while ((i > 0) && (sgn*cmp(v,h[j=(i-1)/2]) > 0)) { h[i] = h[j]; i = j; }
        h[i] = v;
      }
    /* Pop them out in reverse order and store them at the end: */
    while (m > 1) 
      { /* Remove the root: */
        int w = h[0]; 
        int i = 0;  /* {h[i]} is a vacant slot. */
        /* Promote children into vacancy {h[i]} until it reaches the fringe: */
        int ja = 1; /* {h[ja]} is the first child of {h[i]}. */
        while (ja < m)
          { /* Find largest child {h[j]} of {h[i]}: */
            int jb = ja + 1; /*  {h[jb]} is the second child of {h[i]}. */
            int j = ((jb < m) && (sgn*cmp(h[ja],h[jb]) < 0) ? jb : ja);
            /* Promote largest child into hole: */
            h[i] = h[j]; i = j; ja = 2*i + 1;
          }
        m--;
        if (i < m)
          { /* The vacancy did not end up at {h[m]}, so fill it with {h[m]}: */        
            int v = h[m], j; 
            /* Bubble it up to the proper place: */
            while ((i > 0) && (sgn*cmp(v,h[j=(i-1)/2]) > 0)) { h[i] = h[j]; i = j; }
            h[i] = v;
          }
        /* Add the removed element to the sorted region: */
        h[m] = w;
      }
  }
